"""
Miscellaneous utility functions.
"""
from __future__ import annotations
from typing_extensions import TypeIs
import numpy as np
from numpy.typing import NDArray


def _is_list_of_lists(arr: list[float] | list[list[float]]) -> TypeIs[list[list[float]]]:
    return isinstance(arr[0], list)


def masked_array_from_list(
    arr: list[float] | list[list[float]], fill_value: str = "x"
) -> NDArray:
    """Generate a (masked) array from a 1D or 2D list whose elements may contain a fill
    value."""

    is_2D = False
    if _is_list_of_lists(arr):
        is_2D = True
        n_rows = len(arr)
        arr = [item for row in arr for item in row]

    data = np.empty(len(arr))
    mask = np.full(len(arr), False)
    has_mask = False
    for idx, i in enumerate(arr):
        if i == fill_value:
            mask[idx] = True
            has_mask = True
        else:
            data[idx] = i
    if has_mask:
        out = np.ma.masked_array(data, mask=mask)
    else:
        out = data
    if is_2D:
        out = out.reshape(n_rows, -1, order="C")
    return out



from pathlib import Path

from matflow.param_classes import LoadCase
from matflow.tests.utils import make_test_data_YAML_workflow


def test_single_multistep_uniaxial():
    args = {
        "num_increments": 200,
        "total_time": 100,
        "direction": "x",
        "target_def_grad_rate": 1e-3,
    }
    lc1 = LoadCase.uniaxial(**args)
    lc2 = LoadCase.multistep(steps=[{"type": "uniaxial", **args}])
    assert lc1 == lc2


def test_load_case_yaml_init(null_config, tmp_path: Path, load_case_1: LoadCase):
    wk = make_test_data_YAML_workflow("define_load.yaml", path=tmp_path)
    load_case = wk.tasks.define_load_case.elements[0].inputs.load_case.value
    assert load_case == load_case_1



import numpy as np
from numpy.typing import NDArray
from scipy.stats import norm


def update_proposal(lambda_, all_accept, all_x, all_g):
    """Adaptively update the proposal variance according to the acceptance rate of one or
    more Markov chains so far.
    """

    if all_g is not None:
        # stack from multiple chains (elements):
        all_x = np.array([i[:] for i in all_x])
        all_g = np.vstack([i[:] for i in all_g])
        all_accept = np.vstack([i[:] for i in all_accept])

    lambda_latest_iter = sorted(lambda_.items(), key=lambda x: int(x[0].split("_")[1]))[
        -1
    ][1]
    batch_idx = lambda_latest_iter["loop_idx"]["proposal_update"]
    lambda_ = lambda_latest_iter["value"]

    print(f"{lambda_latest_iter=!r}")
    print(f"{lambda_=!r}")

    if all_accept is None:
        # first iteration
        return {
            "lambda_": lambda_,
            "all_x": all_x,
            "all_g": all_g,
            "all_accept": all_accept,
        }

    A_STAR = 0.44

    num_accepts = all_accept.size
    num_chains_per_update = all_accept.shape[0]
    num_accepts_per_batch = int(num_accepts / batch_idx)
    batch_accept_col_idx = -int(num_accepts_per_batch / num_chains_per_update)

    # TODO: should include initial sample in batch_accept probably?
    batch_accept = all_accept[:, batch_accept_col_idx:]
    batch_accept_mean = np.mean(batch_accept)

    lambda_ = lambda_ * np.exp(
        (1 / np.sqrt(batch_idx + 1)) * (batch_accept_mean - A_STAR)
    )

    return {"lambda_": lambda_, "all_x": all_x, "all_g": all_g, "all_accept": all_accept}
